#
# Copyright (c) SAS Institute Inc.
# SPDX-License-Identifier: Apache-2.0
# this has been significantly modified from the original#
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.


import unittest

from univers.maven import Version
from univers.maven import VersionRange
from univers.maven import RestrictionParseError
from univers.maven import VersionRangeParseError
from univers.maven import Restriction


class TestRestriction(unittest.TestCase):
    def test_everyting_spec(self):
        r = Restriction("[,)")
        assert r.lower_bound is None
        assert r.lower_bound_inclusive
        assert r.upper_bound is None
        assert not r.upper_bound_inclusive

        assert "0.1.0" in r
        assert "1.0" in r
        assert "2.0" in r

    def test_exclusive_lower_bound(self):
        r = Restriction("(1.0,2.0]")
        assert str(r.lower_bound) == "1.0"
        assert not r.lower_bound_inclusive
        assert str(r.upper_bound) == "2.0"
        assert r.upper_bound_inclusive

        assert "0.1" not in r
        assert "1.0" not in r
        assert "1.1" in r

    def test_no_lower_limit(self):
        r = Restriction("(,1.0]")
        assert r.lower_bound is None
        assert not r.lower_bound_inclusive
        assert str(r.upper_bound) == "1.0"
        assert r.upper_bound_inclusive
        assert "0.1" in r
        assert "1.0" in r
        assert "1.1" not in r
        assert "2.0" not in r

    def test_inclusive_range(self):
        r = Restriction("[1.0]")
        assert str(r.lower_bound) == "1.0"
        assert r.lower_bound_inclusive
        assert str(r.upper_bound) == "1.0"
        assert r.upper_bound_inclusive

        assert "0.8" not in r
        assert "1.0" in r
        assert "1.1" not in r
        assert "2.0" not in r

        r = Restriction("[1.2,1.3]")
        assert str(r.lower_bound) == "1.2"
        assert r.lower_bound_inclusive
        assert str(r.upper_bound) == "1.3"
        assert r.upper_bound_inclusive

        assert "0.8" not in r
        assert "1.1" not in r
        assert "1.2" in r
        assert "1.2.1" in r
        assert "1.3" in r
        assert "1.3.1" not in r
        assert "2.0" not in r

    def test_exclusive_upper_bound(self):
        r = Restriction("[1.0,2.0)")
        assert str(r.lower_bound) == "1.0"
        assert r.lower_bound_inclusive
        assert str(r.upper_bound) == "2.0"
        assert not r.upper_bound_inclusive

        assert "0.8" not in r
        assert "1.0" in r
        assert "1.9" in r
        assert "2.0" not in r
        assert "3.0" not in r

        r = Restriction("[1.5,)")
        assert str(r.lower_bound) == "1.5"
        assert r.lower_bound_inclusive
        assert r.upper_bound is None
        assert not r.upper_bound_inclusive

        assert "0.8" not in r
        assert "1.4" not in r
        assert "1.5" in r
        assert "100.3" in r

    def test_invalid_restrictions(self):
        tests = (
            "(1.0)",
            "[1.0)",
            "(1.0]",
            "(1.0,1.0]",
            "[1.0,1.0)",
            "(1.0,1.0)",
            "[1.1,1.0]",
        )
        for spec in tests:
            self.assertRaises(
                RestrictionParseError,
                Restriction,
                spec,
            )

    def test_compare(self):
        r1 = Restriction("(1.0,2.0]")

        assert r1 == r1
        assert r1 == "(1.0,2.0]"
        assert 1 < r1

    def test_string_repr(self):
        for test in (
            "[1.0]",
            "[1.0,)",
            "[1.0,2.0]",
            "[1.0,2.0)",
            "(1.0,2.0)",
            "[,2.0]",
            "[,2.0)",
            "(,2.0)",
        ):
            assert str(Restriction(test)) == test


class TestVersion(unittest.TestCase):
    """Tests the Version object"""

    def _assert_version_equal(self, v1, v2):
        V1 = Version(v1)
        V2 = Version(v2)
        # test Version equal Version
        assert V1 == V2, "%s != %s" % (V1, V2)
        assert V2 == V1, "%s != %s" % (V2, V1)

        # test str equal Version
        assert v1 == V2, "%s != %s" % (v1, V2)
        assert v2 == V1, "%s != %s" % (v2, V1)

        # test Version equal str
        assert V1 == v2, "%s != %s" % (V1, v2)
        assert V2 == v1, "%s != %s" % (V2, v1)

    def _assert_version_order(self, v1, v2):
        V1 = Version(v1)
        V2 = Version(v2)

        # Version <-> Version order
        assert V1 < V2, "%s >= %s" % (V1._parsed, V2._parsed)
        assert V2 > V1, "%s >= %s" % (V2._parsed, V1._parsed)

        # Version <-> str order
        assert V1 < v2, "%s >= %s" % (V1._parsed, v2)
        assert V2 > v1, "%s >= %s" % (V2._parsed, v1)

        # str <-> Version order
        assert v1 < V2, "%s >= %s" % (v1, V2._parsed)
        assert v2 > V1, "%s >= %s" % (v2, V1._parsed)

    def test_from_string(self):
        test_pairs = (
            # weird versions
            (".1", (0, 1)),
            ("-1", ((1,),)),
            # test some major.minor.tiny parsing
            ("1", (1,)),
            ("1.0", (1,)),
            ("1.0.0", (1,)),
            ("1.0.0.0", (1,)),
            ("11", (11,)),
            ("11.0", (11,)),
            ("1-1", (1, (1,))),
            ("1-1-1", (1, (1, (1,)))),
            (" 1 ", (1,)),
            # test qualifeirs
            ("1.0-ALPHA", (1, ("alpha",))),
            ("1-alpha", (1, ("alpha",))),
            ("1.0ALPHA", (1, ("alpha",))),
            ("1-alpha", (1, ("alpha",))),
            ("1.0-A", (1, ("a",))),
            ("1-a", (1, ("a",))),
            ("1.0A", (1, ("a",))),
            ("1a", (1, ("a",))),
            ("1.0-BETA", (1, ("beta",))),
            ("1-beta", (1, ("beta",))),
            ("1.0-B", (1, ("b",))),
            ("1-b", (1, ("b",))),
            ("1.0B", (1, ("b",))),
            ("1b", (1, ("b",))),
            ("1.0-MILESTONE", (1, ("milestone",))),
            ("1.0-milestone", (1, ("milestone",))),
            ("1-M", (1, ("m",))),
            ("1.0-m", (1, ("m",))),
            ("1M", (1, ("m",))),
            ("1m", (1, ("m",))),
            ("1.0-RC", (1, ("rc",))),
            ("1-rc", (1, ("rc",))),
            ("1.0-SNAPSHOT", (1, ("snapshot",))),
            ("1.0-snapshot", (1, ("snapshot",))),
            ("1-SP", (1, ("sp",))),
            ("1.0-sp", (1, ("sp",))),
            ("1-GA", (1,)),
            ("1-ga", (1,)),
            ("1.0-FINAL", (1,)),
            ("1-final", (1,)),
            ("1.0-CR", (1, ("rc",))),
            ("1-cr", (1, ("rc",))),
            # test some transition
            ("1.0-alpha1", (1, ("alpha", (1,)))),
            ("1.0-alpha2", (1, ("alpha", (2,)))),
            ("1.0.0alpha1", (1, ("alpha", (1,)))),
            ("1.0-beta1", (1, ("beta", (1,)))),
            ("1-beta2", (1, ("beta", (2,)))),
            ("1.0.0beta1", (1, ("beta", (1,)))),
            ("1.0-BETA1", (1, ("beta", (1,)))),
            ("1-BETA2", (1, ("beta", (2,)))),
            ("1.0.0BETA1", (1, ("beta", (1,)))),
            ("1.0-milestone1", (1, ("milestone", (1,)))),
            ("1.0-milestone2", (1, ("milestone", (2,)))),
            ("1.0.0milestone1", (1, ("milestone", (1,)))),
            ("1.0-MILESTONE1", (1, ("milestone", (1,)))),
            ("1.0-milestone2", (1, ("milestone", (2,)))),
            ("1.0.0MILESTONE1", (1, ("milestone", (1,)))),
            ("1.0-alpha2snapshot", (1, ("alpha", (2, ("snapshot",))))),
        )

        for test, expected in test_pairs:
            v = Version(test)
            assert v._parsed == expected, "Version(%s) == %s, want %s" % (test, v._parsed, expected)

    def test_version_qualifiers(self):
        version_qualifiers = (
            "1-alpha2snapshot",
            "1-alpha2",
            "1-alpha-123",
            "1-beta-2",
            "1-beta123",
            "1-m2",
            "1-m11",
            "1-rc",
            "1-cr2",
            "1-rc123",
            "1-SNAPSHOT",
            "1",
            "1-sp",
            "1-sp2",
            "1-sp123",
            "1-abc",
            "1-def",
            "1-pom-1",
            "1-1-snapshot",
            "1-1",
            "1-2",
            "1-123",
        )
        for idx, low in enumerate(version_qualifiers[:-1]):
            for high in version_qualifiers[idx + 1 :]:
                self._assert_version_order(low, high)

    def test_version_numbers(self):
        version_numbers = (
            "2.0",
            "2-1",
            "2.0.a",
            "2.0.0.a",
            "2.0.2",
            "2.0.123",
            "2.1.0",
            "2.1-a",
            "2.1b",
            "2.1-x",
            "2.1-1",
            "2.1.0.1",
            "2.2",
            "2.123",
            "11.a2",
            "11.a11",
            "11.b2",
            "11.b11",
            "11.m2",
            "11.m11",
            "11",
            "11.a",
            "11b",
            "11c",
            "11m",
        )
        for idx, low in enumerate(version_numbers[:-1]):
            for high in version_numbers[idx + 1 :]:
                self._assert_version_order(low, high)

        unicode_version_numbers = (
            # again, but with unicode input
            "2.0",
            "2-1",
            "2.0.a",
            "2.0.0.a",
            "2.0.2",
            "2.0.123",
            "2.1.0",
            "2.1-a",
            "2.1b",
            "2.1-x",
            "2.1-1",
            "2.1.0.1",
            "2.2",
            "2.123",
            "11.a2",
            "11.a11",
            "11.b2",
            "11.b11",
            "11.m2",
            "11.m11",
            "11",
            "11.a",
            "11b",
            "11c",
            "11m",
        )
        for idx, low in enumerate(unicode_version_numbers[:-1]):
            for high in unicode_version_numbers[idx + 1 :]:
                self._assert_version_order(low, high)

    def test_version_equality(self):
        self._assert_version_equal("1", "1")
        self._assert_version_equal("1", "1.0")
        self._assert_version_equal("1", "1.0.0")
        self._assert_version_equal("1.0", "1.0.0")
        self._assert_version_equal("1", "1-0")
        self._assert_version_equal("1", "1.0-0")
        self._assert_version_equal("1.0", "1.0-0")
        # no separator between number and character
        self._assert_version_equal("1a", "1-a")
        self._assert_version_equal("1a", "1.0-a")
        self._assert_version_equal("1a", "1.0.0-a")
        self._assert_version_equal("1.0a", "1-a")
        self._assert_version_equal("1.0.0a", "1-a")
        self._assert_version_equal("1x", "1-x")
        self._assert_version_equal("1x", "1.0-x")
        self._assert_version_equal("1x", "1.0.0-x")
        self._assert_version_equal("1.0x", "1-x")
        self._assert_version_equal("1.0.0x", "1-x")

        # aliases
        self._assert_version_equal("1ga", "1")
        self._assert_version_equal("1final", "1")
        self._assert_version_equal("1cr", "1rc")

        # special "aliases" a, b and m for alpha, beta and milestone
        self._assert_version_equal("1a1", "1-alpha-1")
        self._assert_version_equal("1b2", "1-beta-2")
        self._assert_version_equal("1m3", "1-milestone-3")

        # case insensitive
        self._assert_version_equal("1X", "1x")
        self._assert_version_equal("1A", "1a")
        self._assert_version_equal("1B", "1b")
        self._assert_version_equal("1M", "1m")
        self._assert_version_equal("1Ga", "1")
        self._assert_version_equal("1GA", "1")
        self._assert_version_equal("1Final", "1")
        self._assert_version_equal("1FinaL", "1")
        self._assert_version_equal("1FINAL", "1")
        self._assert_version_equal("1Cr", "1Rc")
        self._assert_version_equal("1cR", "1rC")
        self._assert_version_equal("1m3", "1Milestone3")
        self._assert_version_equal("1m3", "1MileStone3")
        self._assert_version_equal("1m3", "1MILESTONE3")

        # unicode
        self._assert_version_equal("1", "1")
        self._assert_version_equal("1", "1.0")
        self._assert_version_equal("1", "1.0.0")
        self._assert_version_equal("1.0", "1.0.0")
        self._assert_version_equal("1", "1-0")
        self._assert_version_equal("1", "1.0-0")
        self._assert_version_equal("1.0", "1.0-0")

        self._assert_version_equal("1", "1")
        self._assert_version_equal("1", "1.0")
        self._assert_version_equal("1", "1.0.0")
        self._assert_version_equal("1.0", "1.0.0")
        self._assert_version_equal("1", "1-0")
        self._assert_version_equal("1", "1.0-0")
        self._assert_version_equal("1.0", "1.0-0")

        self._assert_version_equal("1", "1")
        self._assert_version_equal("1", "1.0")
        self._assert_version_equal("1", "1.0.0")
        self._assert_version_equal("1.0", "1.0.0")
        self._assert_version_equal("1", "1-0")
        self._assert_version_equal("1", "1.0-0")
        self._assert_version_equal("1.0", "1.0-0")

    def test_version_compare(self):
        self._assert_version_order("1", "2")
        self._assert_version_order("1.5", "2")
        self._assert_version_order("1", "2.5")
        self._assert_version_order("1.0", "1.1")
        self._assert_version_order("1.1", "1.2")
        self._assert_version_order("1.0.0", "1.1")
        self._assert_version_order("1.0.1", "1.1")
        self._assert_version_order("1.1", "1.2.0")
        self._assert_version_order("1.0-alpha-1", "1.0")
        self._assert_version_order("1.0-alpha-1", "1.0-alpha-2")
        self._assert_version_order("1.0-alpha-1", "1.0-beta-1")
        self._assert_version_order("1.0-beta-1", "1.0-SNAPSHOT")
        self._assert_version_order("1.0-SNAPSHOT", "1.0")
        self._assert_version_order("1.0-alpha-1-SNAPSHOT", "1.0-alpha-1")
        self._assert_version_order("1.0", "1.0-1")
        self._assert_version_order("1.0-1", "1.0-2")
        self._assert_version_order("1.0.0", "1.0-1")
        self._assert_version_order("2.0-1", "2.0.1")
        self._assert_version_order("2.0.1-klm", "2.0.1-lmn")
        self._assert_version_order("2.0.1", "2.0.1-xyz")
        self._assert_version_order("2.0.1", "2.0.1-123")
        self._assert_version_order("2.0.1-xyz", "2.0.1-123")
        # unicode input
        self._assert_version_order("1", "2")
        self._assert_version_order("1.5", "2")
        self._assert_version_order("1", "2.5")
        self._assert_version_order("1.0", "1.1")
        self._assert_version_order("1.1", "1.2")
        self._assert_version_order("1.0.0", "1.1")
        self._assert_version_order("1.0.1", "1.1")
        self._assert_version_order("1.1", "1.2.0")
        self._assert_version_order("1.0-alpha-1", "1.0")
        self._assert_version_order("1.0-alpha-1", "1.0-alpha-2")
        self._assert_version_order("1.0-alpha-1", "1.0-beta-1")
        self._assert_version_order("1.0-beta-1", "1.0-SNAPSHOT")
        self._assert_version_order("1.0-SNAPSHOT", "1.0")
        self._assert_version_order("1.0-alpha-1-SNAPSHOT", "1.0-alpha-1")
        self._assert_version_order("1.0", "1.0-1")
        self._assert_version_order("1.0-1", "1.0-2")
        self._assert_version_order("1.0.0", "1.0-1")
        self._assert_version_order("2.0-1", "2.0.1")
        self._assert_version_order("2.0.1-klm", "2.0.1-lmn")
        self._assert_version_order("2.0.1", "2.0.1-xyz")
        self._assert_version_order("2.0.1", "2.0.1-123")
        self._assert_version_order("2.0.1-xyz", "2.0.1-123")

        self._assert_version_order("1", "2")
        self._assert_version_order("1.5", "2")
        self._assert_version_order("1", "2.5")
        self._assert_version_order("1.0", "1.1")
        self._assert_version_order("1.1", "1.2")
        self._assert_version_order("1.0.0", "1.1")
        self._assert_version_order("1.0.1", "1.1")
        self._assert_version_order("1.1", "1.2.0")
        self._assert_version_order("1.0-alpha-1", "1.0")
        self._assert_version_order("1.0-alpha-1", "1.0-alpha-2")
        self._assert_version_order("1.0-alpha-1", "1.0-beta-1")
        self._assert_version_order("1.0-beta-1", "1.0-SNAPSHOT")
        self._assert_version_order("1.0-SNAPSHOT", "1.0")
        self._assert_version_order("1.0-alpha-1-SNAPSHOT", "1.0-alpha-1")
        self._assert_version_order("1.0", "1.0-1")
        self._assert_version_order("1.0-1", "1.0-2")
        self._assert_version_order("1.0.0", "1.0-1")
        self._assert_version_order("2.0-1", "2.0.1")
        self._assert_version_order("2.0.1-klm", "2.0.1-lmn")
        self._assert_version_order("2.0.1", "2.0.1-xyz")
        self._assert_version_order("2.0.1", "2.0.1-123")
        self._assert_version_order("2.0.1-xyz", "2.0.1-123")

        self._assert_version_order("1", "2")
        self._assert_version_order("1.5", "2")
        self._assert_version_order("1", "2.5")
        self._assert_version_order("1.0", "1.1")
        self._assert_version_order("1.1", "1.2")
        self._assert_version_order("1.0.0", "1.1")
        self._assert_version_order("1.0.1", "1.1")
        self._assert_version_order("1.1", "1.2.0")
        self._assert_version_order("1.0-alpha-1", "1.0")
        self._assert_version_order("1.0-alpha-1", "1.0-alpha-2")
        self._assert_version_order("1.0-alpha-1", "1.0-beta-1")
        self._assert_version_order("1.0-beta-1", "1.0-SNAPSHOT")
        self._assert_version_order("1.0-SNAPSHOT", "1.0")
        self._assert_version_order("1.0-alpha-1-SNAPSHOT", "1.0-alpha-1")
        self._assert_version_order("1.0", "1.0-1")
        self._assert_version_order("1.0-1", "1.0-2")
        self._assert_version_order("1.0.0", "1.0-1")
        self._assert_version_order("2.0-1", "2.0.1")
        self._assert_version_order("2.0.1-klm", "2.0.1-lmn")
        self._assert_version_order("2.0.1", "2.0.1-xyz")
        self._assert_version_order("2.0.1", "2.0.1-123")
        self._assert_version_order("2.0.1-xyz", "2.0.1-123")

    def test_compare(self):
        assert 1 < Version("1.0")
        assert VersionRange("1.0") == Version("1.0")

    def test_compare_errors(self):
        v = Version("1.0")
        self.assertRaises(RuntimeError, v._compare, v, 1.0)
        self.assertRaises(RuntimeError, v._int_compare, v, 1.0)
        self.assertRaises(RuntimeError, v._list_compare, v, 1.0)
        self.assertRaises(RuntimeError, v._string_compare, v, 1.0)


class TestVersionRange(unittest.TestCase):
    def test_no_lower_limit(self):
        vr = VersionRange("(,1.0]")
        assert len(vr.restrictions) == 1
        assert vr.version is None

    def test_single_spec(self):
        vr = VersionRange("1.0")
        assert len(vr.restrictions) == 1
        assert str(vr.version) == "1.0"

    def test_inclusive_range(self):
        vr = VersionRange("[1.0]")
        assert len(vr.restrictions) == 1
        assert vr.version is None

        vr = VersionRange("[1.2,1.3]")
        assert len(vr.restrictions) == 1
        assert vr.version is None

    def test_exclusive_upper_bound(self):
        vr = VersionRange("[1.0,2.0)")
        assert len(vr.restrictions) == 1
        assert vr.version is None

        vr = VersionRange("[1.5,)")
        assert len(vr.restrictions) == 1
        assert vr.version is None

    def test_multiple_restrictions(self):
        vr = VersionRange("(,1.0],[1.2,)")
        assert len(vr.restrictions) == 2
        assert vr.version is None

    def test_snapshots(self):
        vr = VersionRange("[1.0,)")
        assert "1.0-SNAPSHOT" not in vr

        vr = VersionRange("[1.0,1.1-SNAPSHOT]")
        assert "1.1-SNAPSHOT" in vr

        vr = VersionRange("[1.0,1.2]")
        assert "1.0-SNAPSHOT" not in vr
        assert "1.1-SNAPSHOT" in vr
        assert "1.2-SNAPSHOT" in vr
        assert "1.3-SNAPSHOT" not in vr

        vr = VersionRange("[1.0,1.2-SNAPSHOT]")
        assert "1.1-SNAPSHOT" in vr
        assert "1.2-SNAPSHOT" in vr

        vr = VersionRange("[1.0-SNAPSHOT,1.2]")
        assert "1.0-SNAPSHOT" in vr
        assert "1.1-SNAPSHOT" in vr

        vr = VersionRange("1.0-SNAPSHOT")
        assert "1.0-SNAPSHOT" in vr

        vr = VersionRange("[0.1,1.0)")
        assert "0.1" in vr
        assert "0.1.0" in vr
        assert "0.1-SNAPSHOT" not in vr
        assert "0.1.0-SNAPSHOT" not in vr

    def test_long_version(self):
        vr = VersionRange("[5.0.9.0,5.0.10.0)")
        assert Version("5.0.9.0") in vr
        assert "5.0.9.0" in vr

    def test_contains(self):
        test_pairs = (
            ("2.0.5", True),
            ("2.0.4", True),
            ("[2.0.5]", True),
            ("[2.0.6,)", False),
            ("[2.0.6]", False),
            ("2.0,2.1]", True),
            ("[2.0,2.0.3]", False),
            ("[2.0,2.0.5]", True),
            ("[2.0,2.0.5)", False),
        )
        v = Version("2.0.5")
        for spec, expected in test_pairs:
            vr = VersionRange(spec)
            assert (v in vr) == expected
            assert ("2.0.5" in vr) == expected

        assert VersionRange("2.0.5") in VersionRange("[2.0,3.0)")

        assert VersionRange("[2.0.5]") not in VersionRange("[2.0,3.0)")

    def test_invalid_ranges(self):
        for spec in (
            "(1.0,2.0",
            "[1.0,1.2),1.3",
            "[1.0,1.2),(1.1,1.3]",
            "[1.1,1.3),(1.0,1.2]",
            "(1.1,1.2],[1.0,1.1)",
        ):
            self.assertRaises(VersionRangeParseError, VersionRange, spec)

    def test_compare(self):
        vr1 = VersionRange("(1.0,2.0]")

        assert vr1 == vr1
        assert vr1 == "(1.0,2.0]"
        assert 1 < vr1

    def test_str(self):
        for inp in ("[1.0,2.0]", "1.0"):
            actual = str(VersionRange(inp))
            assert inp == actual, "VersionRange(%s) == %s, wanted %s" % (inp, actual, inp)

    def test_fromversion(self):
        v = Version("1.0")
        vr = VersionRange.from_version(v)

        assert vr.version == v

    def test_match_versions(self):
        versions = [Version("0.1"), Version("1.0"), Version("1.1"), Version("2.0"), Version("2.1")]
        vr = VersionRange("(1.0,2.0]")
        assert vr.match_version(versions) == "2.0"
        assert vr.match_version(versions[:3]) == "1.1"



#
# Copyright (c) Chad Fowler, Rich Kilmer, Jim Weirich and others.
# Portions copyright (c) Engine Yard and Andre ArkoFacebook, Inc. and its affiliates.
#
# SPDX-License-Identifier: MIT
#
# Originally from https://github.com/rubygems/rubygems

from univers.gem import GemRequirement


def test_satisfied_by():

    assert not GemRequirement("!= 1").satisfied_by("1")
    assert GemRequirement("!= 1").satisfied_by("2")

    assert not GemRequirement("!= 1", "= 2").satisfied_by("1")
    assert GemRequirement("!= 1", "= 2").satisfied_by("2")

    assert not GemRequirement("!= 1", "> 1").satisfied_by("1")
    assert GemRequirement("!= 1", "> 1").satisfied_by("2")

    assert not GemRequirement("!= 1", ">= 1").satisfied_by("1")
    assert GemRequirement("!= 1", ">= 1").satisfied_by("2")

    assert not GemRequirement("= 1", ">= 0.1", "<= 1.1").satisfied_by("0.2")
    assert GemRequirement("= 1", ">= 0.1", "<= 1.1").satisfied_by("1")
    assert not GemRequirement("= 1", ">= 0.1", "<= 1.1").satisfied_by("3")

    assert GemRequirement("= 1", ">= 1", "<= 1").satisfied_by("1")
    assert not GemRequirement("= 1", ">= 1", "<= 1").satisfied_by("2")
    assert not GemRequirement("= 1", ">= 1", "<= 1").satisfied_by("0.1")

    assert GemRequirement("= 1", "~> 1").satisfied_by("1")
    assert not GemRequirement("= 1", "~> 1").satisfied_by("1.1")

    assert GemRequirement(">= 0.z", "= 0").satisfied_by("0")
    assert not GemRequirement(">= 0.z", "= 0").satisfied_by("1")
    assert not GemRequirement(">= 0.z", "= 0").satisfied_by("0.1")
    assert not GemRequirement(">= 0.z", "= 0").satisfied_by("0.z")

    assert GemRequirement(">= 0").satisfied_by("1")
    assert GemRequirement(">= 0").satisfied_by("2")
    assert GemRequirement(">= 0").satisfied_by("0")

    assert not GemRequirement(">= 1.0.0", "< 2.0.0").satisfied_by("3")
    assert GemRequirement(">= 1.0.0", "< 2.0.0").satisfied_by("1.5.1")

    assert GemRequirement("~> 1").satisfied_by("1")
    assert GemRequirement("~> 1").satisfied_by("1.1")
    assert not GemRequirement("~> 1").satisfied_by("2")

    assert not GemRequirement("~> 2.0", "~> 2.1").satisfied_by("1")
    assert GemRequirement("~> 2.0", "~> 2.1").satisfied_by("2.1.2")

    assert not GemRequirement(">= 4.1.0", "< 5.0", "= 5.2.1").satisfied_by("1")
    assert not GemRequirement(">= 4.1.0", "< 5.0", "= 5.2.1").satisfied_by("5.2.1")
    assert not GemRequirement(
        "< 5.0",
        "< 5.3",
        "< 6.0",
        "< 6",
        "= 5.2.0",
        "> 2",
        ">= 3.0",
        ">= 3.1",
        ">= 3.2",
        ">= 4.0.0",
        ">= 4.1.0",
        ">= 4.2.0",
        ">= 4.2",
        ">= 4",
    ).satisfied_by("5.2.0")
    assert not GemRequirement("!= 1", "< 2", "> 2").satisfied_by("1")
    assert not GemRequirement("!= 1", "<= 1", ">= 1").satisfied_by("1")
    assert not GemRequirement("< 2", "> 2").satisfied_by("1")
    assert not GemRequirement("< 2", "> 2", "= 2").satisfied_by("1")
    assert not GemRequirement("= 1", "!= 1").satisfied_by("1")
    assert not GemRequirement("= 1", "= 2").satisfied_by("1")
    assert not GemRequirement("= 1", "~> 2").satisfied_by("1")
    assert not GemRequirement(">= 0", "<= 0.a").satisfied_by("1")
    assert not GemRequirement("~> 2.0", "~> 3").satisfied_by("1")



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import operator
from functools import total_ordering

import attr
from univers.utils import remove_spaces
from univers.versions import Version

try:
    # only stanadard in Python 3.10 and up
    from itertools import pairwise  # NOQA
except ImportError:
    # back from docs at https://docs.python.org/3/library/itertools.html#itertools.pairwise
    import itertools

    def pairwise(iterable):
        a, b = itertools.tee(iterable)
        next(b, None)
        return zip(a, b)


"""
Universal version constraint object that stores a comparator such as "=" or "<="
and a Version object using a class that is specific to a package type (aka.
ecosystem)
"""


def operator_star(a, b):
    """
    Comparison operator for the star "*" constraint comparator. Since it matches
    any version, it is always True.
    """
    return True


# note: ORDER MATTER here: we tests startswith(key) for each key in sequence
COMPARATORS = {
    ">=": operator.ge,
    "<=": operator.le,
    "!=": operator.ne,
    "<": operator.lt,
    ">": operator.gt,
    "=": operator.eq,
    "*": operator_star,
}


@total_ordering
@attr.s(frozen=True, repr=True, str=False, order=False, eq=True, hash=True)
class VersionConstraint:
    """
    Represent a single constraint composed of a comparator and a version.

    VersionConstraint is:
    - comparable and orderable e.g., implements functools.total_ordering
      by version then comparator.
    - immutable and hashable.
    """

    # one of the COMPARATORS
    comparator = attr.ib(type=str, default="=")

    # a Version subclass instance
    version = attr.ib(type=Version, default=None)

    # a function for the comparator
    comp_operator = attr.ib(default=None, repr=False)

    # a Version subclass
    version_class = attr.ib(type=Version, default=None, repr=False)

    def __attrs_post_init__(self):
        # Notes: setattr is used because this is an immutable frozen instance.
        # See https://www.attrs.org/en/stable/init.html?#post-init
        try:
            object.__setattr__(self, "comp_operator", COMPARATORS[self.comparator])
        except KeyError as e:
            raise ValueError(f"Unknown comparator: {self.comparator}") from e

        if self.version and not isinstance(self.version, Version):
            raise TypeError(
                f"version must be a 'Version' instance and not: {self.version.__class__!r}"
            )

        if not self.version_class:
            if self.version:
                object.__setattr__(self, "version_class", self.version.__class__)
            else:
                raise ValueError("Cannot build a VersionConstraint without a version class")

    def __str__(self):
        """
        Return a string representing this constraint.
        For example::
        >>> assert str(VersionConstraint(comparator=">=", version=Version("2.3"))) == ">=2.3"
        >>> assert str(VersionConstraint(comparator="*", version_class=Version)) == "*"
        >>> assert str(VersionConstraint(comparator="<", version=Version("2.3"))) == "<2.3"
        >>> assert str(VersionConstraint(comparator="=", version=Version("2.3.0"))) == "2.3.0"
        >>> assert str(VersionConstraint(version=Version("2.3.0"))) == "2.3.0"
        """
        if self.comparator == "*":
            return "*"

        if self.comparator == "=":
            return str(self.version)

        version = str(self.version)
        return f"{self.comparator}{version}"

    to_string = __str__

    def to_dict(self):
        return dict(comparator=self.comparator, version=str(self.version))

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.comparator == other.comparator and self.version == other.version

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        # we compare tuples, version first
        return (self.version, self.comparator).__lt__((other.version, other.comparator))

    @classmethod
    def from_string(cls, string, version_class):
        """
        Return a single VersionConstraint built from a constraint ``string`` and
        a ``version_class`` Version class.
        """
        constraint_string = remove_spaces(string)

        # A version range specifier contains only printable ASCII letters, digits and
        # punctuation.
        is_ascii = len(constraint_string) + 2 == len(ascii(constraint_string))
        if not is_ascii:
            raise ValueError(f"Invalid non ASCII characters: {string!r}")

        comparator, version = cls.split(constraint_string)
        if comparator not in COMPARATORS:
            raise ValueError(f"Unknown comparator: {comparator!r}")

        if not version and comparator != "*":
            raise ValueError("Empty version")

        if comparator == "*":
            version = None
        else:
            version = version_class(version)
        return cls(comparator=comparator, version=version, version_class=version_class)

    @staticmethod
    def split(string):
        """
        Return a tuple of (comparator, version) strings given a
        constraint ``string`` such as ">=2.3".

        For example::
        >>> assert VersionConstraint.split(">=2.3") == (">=", "2.3",)
        >>> assert VersionConstraint.split("  <   =  2 . 3  ") == ("<=", "2.3",)
        >>> assert VersionConstraint.split("2.3") == ("=", "2.3",)
        >>> assert VersionConstraint.split("*2.3") == ("*", "",)
        >>> assert VersionConstraint.split("*") == ("*", "",)
        >>> assert VersionConstraint.split("<2.3") == ("<", "2.3",)
        >>> assert VersionConstraint.split(">2.3") == (">", "2.3",)
        >>> assert VersionConstraint.split("!=2.3") == ("!=", "2.3",)
        """
        constraint_string = remove_spaces(string)

        # special case for star
        if constraint_string.startswith("*"):
            return "*", ""

        for comparator in COMPARATORS:
            if constraint_string.startswith(comparator):
                # NOTE: we do not report an error if this is not valid
                version = constraint_string.lstrip(comparator)
                if comparator == "*":
                    version = ""
                return comparator, version

        # default to equality
        return "=", constraint_string

    def __contains__(self, version):
        """
        Return a True if the ``version`` Version is contained in this
        VersionConstraint or "satisfies" this VersionConstraint.

        For example::
        >>> from univers.versions import PypiVersion
        >>> v22 = PypiVersion("2.2")
        >>> v23 = PypiVersion("2.3")
        >>> v24 = PypiVersion("2.4")
        >>> assert v23 in VersionConstraint(comparator="=", version=v23)
        >>> assert v24 not in VersionConstraint(comparator="=", version=v23)
        >>> try:
        ...   None in VersionConstraint(comparator="=", version=v23)
        ... except ValueError:
        ...   pass

        >>> assert v22 in VersionConstraint(comparator="!=", version=v23)
        >>> assert v23 in VersionConstraint(comparator="!=", version=v24)
        >>> assert v24 not in VersionConstraint(comparator="!=", version=v24)

        >>> assert v24 in VersionConstraint(comparator=">", version=v23)
        >>> assert v23 not in VersionConstraint(comparator=">", version=v23)
        >>> assert v24 in VersionConstraint(comparator=">=", version=v23)
        >>> assert v23 in VersionConstraint(comparator=">=", version=v23)
        >>> assert v22 not in VersionConstraint(comparator=">=", version=v23)

        >>> assert v23 in VersionConstraint(comparator="<", version=v24)
        >>> assert v23 in VersionConstraint(comparator="<=", version=v24)
        >>> assert v24 in VersionConstraint(comparator="<=", version=v24)
        >>> assert v24 not in VersionConstraint(comparator="<", version=v24)
        """
        if not isinstance(version, self.version_class):
            raise ValueError(
                f"Cannot compare {version.__class__!r} instance "
                f"with {self.version_class!r} instance."
            )
        return self.comp_operator(version, self.version)

    contains = __contains__

    @classmethod
    def validate(cls, constraints):
        """
        Raise an assertion error if the ``constraints`` is not a list of
        VersionConstraint objects or if two VersionConstraint contain the same
        Version (ignoring the comparator).
        Also validate that the sequence of comparators is valid.
        Return True otherwise.
        """

        if not isinstance(constraints, (list, tuple)):
            raise ValueError(f"{constraints!r} is a not list or tuple")

        if not all(isinstance(c, VersionConstraint) for c in constraints):
            raise ValueError(f"{constraints!r} can contain only VersionConstraint")

        # Versions are unique. Each ``version`` must be unique in a range and can
        # occur only once in any ``<version-constraint>`` of a range specifier,
        # irrespective of its comparators. Tools must report an error for duplicated
        # versions.
        if len(set(c.version for c in constraints)) != len(constraints):
            raise ValueError(f"{constraints!r} cannot contain duplicated Version")

        # Constraints are sorted by version**. The canonical ordering is the versions
        # order. The ordering of ``<version-constraint>`` is not significant otherwise
        # but this sort order is needed when check if a version is contained in a range.
        constraints.sort()

        return validate_comparators(constraints)

    @classmethod
    def simplify(cls, constraints):
        """
        Return a new simplified ``constraints`` list with duplicated constraints
        removed. This includes removing exact duplicates and redundant
        constraints.
        """
        constraints = deduplicate(constraints)
        constraints = simplify_constraints(constraints)
        return constraints


def deduplicate(constraints):
    """
    Return a new ``constraints`` list with exact duplicated constraints removed.
    """
    seen = set()
    unique = []
    for c in constraints:
        if c not in seen:
            unique.append(c)
            seen.add(c)
    return unique


def validate_comparators(constraints):
    """
    Raise an assertion error if the ``constraints`` list contains an invalid
    sequence of constraint comparators according to ``vers`` rules.
    Return True otherwise.

    The following are the validity rules for contiguous constraints where the
    constraints are canonical e.g., sorted by version and versions are unique
    ignoring comparators:

    - "*" can only occur alone
    - "!=" can be followed by anything, i.e., one of "=", "!=", ">", ">=", "<", "<="

    And ignoring all "!=":
    - "=" can be followed only by one of "=", ">", ">="

    And ignoring all "=" and "!=", there must be an alternation of greater and lesser:
    - "<" and "<=" can only be followed by one of ">", ">="
    - ">" and ">=" can only be followed by one of "<", "<="
    """

    # Starting from a de-duplicated and sorted list of constraints, these extra rules
    # apply to the comparators of any two contiguous constraints to be valid:

    # There is only one star: "*" must only occur once and alone in a range,
    # without any other constraint or version.
    if any(c.comparator == "*" for c in constraints):
        if len(constraints) != 1:
            raise ValueError(f"Invalid {constraints!r}: can contain only one star '*'")
        return True

    # "!=" constraint can be followed by a constraint using any comparator, i.e.,
    # any of "=", "!=", ">", ">=", "<", "<=" as comparator (or no constraint).

    # Ignoring all constraints with "!=" comparators:
    # --> discard != that can occur anywhere
    constraints = [c for c in constraints if c.comparator != "!="]
    if not constraints:
        return True

    # A "=" constraint must be followed only by a constraint with one of "=", ">",
    # ">=" as comparator (or no constraint).
    # --> check that equals is followed only by "=", ">", ">="
    invalid_equal = [
        (cur, nxt)
        for cur, nxt in pairwise(constraints)
        if cur.comparator == "=" and nxt.comparator not in ("=", ">", ">=")
    ]
    if invalid_equal:
        c = "|".join(map(str, constraints))
        i = ", ".join(f"{x}|{y}" for x, y in invalid_equal)
        raise ValueError(
            f"Invalid {c!r}: where {i!r} " "cannot contain an equal = followed by either < or <="
        )

    # And ignoring all constraints with "=" or "!=" comparators:
    # --> discard = that have been validated above
    constraints = [c for c in constraints if c.comparator != "="]
    if not constraints:
        return True

    # the sequence of constraint comparators must be an alternation of greater
    # and lesser comparators:
    # --> from now on this must be an alternation of greater/lesser
    for cur_constraint, nxt_constraint in pairwise(constraints):
        cur_comp = cur_constraint.comparator
        nxt_comp = nxt_constraint.comparator

        # "<" and "<=" must be followed by one of ">", ">=" (or no constraint).
        # ">" and ">=" must be followed by one of "<", "<=" (or no constraint).
        # Tools must report an error for such invalid ranges.
        if (cur_comp in ("<", "<=") and nxt_comp not in (">", ">=")) or (
            cur_comp in (">", ">=") and nxt_comp not in ("<", "<=")
        ):

            raise ValueError(
                f"Invalid {constraints!r}: {cur_constraint!r} "
                f"cannot be followed by {nxt_constraint!r}"
            )

    return True


def simplify_constraints(constraints):
    """
    Return a list of VersionConstraint given a ``constraints`` list by
    discarding redundant constraints according to ``vers`` rules.
    """
    # Start from a list of constraints of comparator and version, sorted by version
    # and where each version occurs only once in any constraint.

    # If the constraints list contains a single constraint (star, equal or anything)
    # return this list and simplification is finished.
    if len(constraints) < 2:
        return constraints

    # Split the constraints list in two sub lists:
    #   a list of "unequal constraints" where the comparator is "!="
    #   a remainder list of "constraints" where the comparator is not "!="
    unequal_constraints = [c for c in constraints if c.comparator == "!="]
    constraints = [c for c in constraints if c.comparator != "!="]

    # If the remainder list of "constraints" is empty, return the "unequal constraints"
    # list and de-duplication is finished.
    if not constraints:
        return unequal_constraints

    # Iterate over the current and next contiguous constraints of this list:
    i = 0
    j = 0

    while i < len(constraints) - 1:
        j = i + 1

        cur = constraints[i]
        nxt = constraints[j]
        cur_comp = cur.comparator
        nxt_comp = nxt.comparator

        # If current comparator is ">" or ">=" and next comparator is "=", ">" or ">=",
        if cur_comp in (">", ">=") and nxt_comp in ("=", ">", ">="):
            # discard next constraint
            constraints.pop(j)

        # If current comparator is "=", "<" or "<="  and next comparator is <" or <=",
        if cur_comp in ("=", "<", "<=") and nxt_comp in ("<", "<="):
            # discard current constraint
            constraints.pop(i)
            # Previous constraint becomes current if if exists.
            if i > 0:
                i -= 1

        # If there is a previous constraint:
        if i > 0:

            prv = constraints[i - 1]
            prv_comp = prv.comparator

            # If previous comparator is ">" or ">=" and current comparator is "=", ">" or ">=",
            if prv_comp in (">", ">=") and cur_comp in ("=", ">", ">="):
                # discard current constraint
                constraints.pop(i)

            # If previous comparator is "=", "<" or "<=" and current comparator is <" or <=",
            if prv_comp in ("=", "<", "<=") and cur_comp in ("<", "<="):
                # discard previous constraint.
                constraints.pop(i - 1)

        i += 1

    # Concatenate the "unequal constraints" list and the filtered "constraints" list
    # Sort by version and return.
    return sorted(set(unequal_constraints + constraints))


def contains_version(version, constraints):
    """
    Return True an assertion error if the ``constraints`` list contains the
    ``version`` Version object according to ``vers`` rules.
    """
    # If the constraint list contains only one item and the comparator is "*",
    # then the "tested version" is IN the range. Check is finished.

    # If the constraint list contains only one item and and the "tested version"
    # satisfies the comparator then the "tested version"  is IN the range.
    # Check is finished.
    if len(constraints) == 1:
        return version in constraints[0]

    # If the "tested version" is equal to the any of the constraint version where
    # the constraint comparator is "=!" then the "tested version" is NOT in the
    # range. Check is finished.
    for constraint in constraints:
        if "!=" in constraint.comparator and version == constraint.version:
            return False

    # If the "tested version" is equal to the any of the constraint version
    # where the constraint comparator is for equality (any of "=", "<=", or ">=")
    # then the "tested version" is in the range. Check is finished.
    for constraint in constraints:
        if "=" in constraint.comparator and version == constraint.version:
            return True

    # Split the constraint list in two sub lists:
    #   a first list where the comparator is "=" or "!="
    #   a second list where the comparator is neither "=" nor "!="
    constraints = [c for c in constraints if c.comparator not in ("=", "!=")]
    if not constraints:
        return False

    # Iterate over the current and next contiguous constraints pairs (aka. pairwise)
    # in the second list.
    # For each current and next constraint:

    cur_comp = nxt_comp = cur_constraint = nxt_constraint = None
    first_iteration = True
    for cur_constraint, nxt_constraint in pairwise(constraints):
        cur_comp = cur_constraint.comparator
        nxt_comp = nxt_constraint.comparator

        # If this is the first iteration and current comparator is "<" or <="
        # and the "tested version" is less than the current version
        # then the "tested version" is IN the range. Check is finished.
        if first_iteration:
            if cur_comp in ("<", "<=") and version < cur_constraint.version:
                return True
            first_iteration = False

        # If current comparator is ">" or >=" and next comparator is "<" or <="
        # and the "tested version" is greater than the current version
        # and the "tested version" is less than the next version
        # then the "tested version" is IN the range. Check is finished.
        if cur_comp in (">", ">=") and nxt_comp in ("<", "<="):
            if version > cur_constraint.version and version < nxt_constraint.version:
                return True

        # If current comparator is "<" or <=" and next comparator is ">" or >="
        # then these versions are out the range. Continue to the next iteration.
        elif cur_comp in ("<", "<=") and nxt_comp in (">", ">="):
            pass

        else:
            # this should never happen as the constraints must be valid going in
            raise Exception(f"Invalid constraints sequence: {constraints }")

    # If this is the last iteration and next comparator is ">" or >="
    # and the "tested version" is greater than the next version
    # then the "tested version" is IN the range. Check is finished.
    if nxt_comp in (">", ">=") and version > nxt_constraint.version:
        return True

    # Reaching here without having finished the check before means that the
    # "tested version" is NOT in the range.
    return False



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

from univers.utils import remove_spaces
from univers.version_constraint import VersionConstraint
from univers.versions import SemverVersion

"""
node-semver and Rubygems semver-like related utilities.
"""


def get_caret_constraints(string):
    """
    Return a tuple of two VersionConstraint of ``SemverVersion`` representing
    the lower and upper bound of version constraint ``string`` that contains a
    caret node-semver- like range. Raise a ValueError if this is not a caret
    range.

    For example:
    >>> lower_bound, upper_bound = get_caret_constraints("^1.0.2")
    >>> vlow = SemverVersion("1.0.2")
    >>> vup = SemverVersion("2.0.0")
    >>> assert lower_bound == VersionConstraint(comparator=">=", version=vlow)
    >>> assert upper_bound == VersionConstraint(comparator="<", version=vup)
    """
    string = remove_spaces(string)
    if not string or not string.startswith("^"):
        raise ValueError(f"Invalid caret version range: {string!r}")

    version = string.lstrip("^")
    lower_bound = SemverVersion(version)
    upper_bound = SemverVersion(str(lower_bound.value.next_major()))

    return (
        VersionConstraint(comparator=">=", version=lower_bound),
        VersionConstraint(comparator="<", version=upper_bound),
    )


def get_tilde_constraints(string, operator="~"):
    """
    Return a tuple of two VersionConstraint of ``SemverVersion`` representing
    the lower and upper bound of a version range ``string`` that contains a
    tilde node-semver-like range.
    Raise a ValueError if this is not a tilde range.

    For example:
    >>> lower_bound, upper_bound = get_tilde_constraints("~1.0.2")
    >>> vlow = SemverVersion("1.0.2")
    >>> vup = SemverVersion("1.1.0")
    >>> assert lower_bound == VersionConstraint(comparator=">=", version=vlow)
    >>> assert upper_bound == VersionConstraint(comparator="<", version=vup)
    """
    string = remove_spaces(string)
    if not string or not string.startswith(operator):
        raise ValueError(f"Invalid version range: {string!r} " f"does not start with {operator!r}")

    version = string.lstrip(operator)
    lower_bound = SemverVersion(version)
    upper_bound = SemverVersion(str(lower_bound.value.next_minor()))

    return (
        VersionConstraint(comparator=">=", version=lower_bound),
        VersionConstraint(comparator="<", version=upper_bound),
    )


# FIXME: this is unlikely correct https://github.com/npm/node-semver/issues/112
def get_pessimistic_constraints(string):
    """
    Return a tuple of two VersionConstraint of ``SemverVersion`` representing
    the lower and upper bound of version range ``string`` that contains a
    pessimistic Ruby range.  Raise a ValueError if this is not a pessimistic
    Rubygems range.

    For example:
    >>> lower_bound, upper_bound = get_pessimistic_constraints("~>2.0.8")
    >>> vlow = SemverVersion("2.0.8")
    >>> vup = SemverVersion("2.1.0")
    >>> assert lower_bound == VersionConstraint(comparator=">=", version=vlow)
    >>> assert upper_bound == VersionConstraint(comparator="<", version=vup)
    """
    return get_tilde_constraints(string, operator="~>")



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

from functools import total_ordering

import attr
import semantic_version
from packaging import version as packaging_version

from univers import arch
from univers import debian
from univers import gem
from univers import gentoo
from univers import maven
from univers import rpm
from univers.utils import remove_spaces

"""
Version classes encapsulating the details of each version syntax.
For instance semver is a version syntax. Python and Debian use another syntax.

Each subclass primary responsibility to is be comparable and orderable
"""

# TODO: Add mozilla versions https://github.com/mozilla-releng/mozilla-version
# TODO: Add conda versions https://github.com/conda/conda/blob/master/conda/models/version.py
#       and https://docs.conda.io/projects/conda-build/en/latest/resources/package-spec.html#build-version-spec


class InvalidVersion(ValueError):
    pass


def is_valid_alpine_version(s):
    """
    Return True is the string `s` is a valid Alpine version.
    We do not support yet version strings that start with
    non-significant zeros.
    For example:
    >>> is_valid_alpine_version("006")
    False
    >>> is_valid_alpine_version("1.2.3")
    True
    >>> is_valid_alpine_version("02-r1")
    False
    """
    left, _, _ = s.partition(".")
    # hanlde the suffix case
    left, _, _ = left.partition("-")
    if not left.isdigit():
        return True
    i = int(left)
    return str(i) == left


@attr.s(frozen=True, order=False, hash=True)
class Version:
    """
    Base version mixin to subclass for each version syntax implementation.

    Each version subclass is:

    - immutable and hashable
    - comparable and orderable e.g., such as implementing all rich comparison
      operators or implementing functools.total_ordering. The default is to
      compare the value as-is.
    """

    # the original string used to build this Version
    string = attr.ib(type=str)

    # the normalized string for this Version, stored without spaces and
    # lowercased. Any leading v is removed too.
    normalized_string = attr.ib(type=str, default=None, repr=False)

    # a comparable scheme-specific version object constructed from the version string
    value = attr.ib(default=None, repr=False)

    def __attrs_post_init__(self):
        normalized_string = self.normalize(self.string)
        if not self.is_valid(normalized_string):
            raise InvalidVersion(f"{self.string!r} is not a valid {self.__class__!r}")

        # Set the normalized string as default value

        # Notes: setattr is used because this is an immutable frozen instance.
        # See https://www.attrs.org/en/stable/init.html?#post-init
        object.__setattr__(self, "normalized_string", normalized_string)
        value = self.build_value(normalized_string)
        object.__setattr__(self, "value", value)

    @classmethod
    def is_valid(cls, string):
        """
        Return True if the ``string`` is a valid version for its scheme or False
        if not valid. The empty string, None, False and 0 are considered invalid.
        Subclasses should implement this.
        """
        return bool(string)

    @classmethod
    def normalize(cls, string):
        """
        Return a normalized version string from ``string ``. Subclass can override.
        """
        # FIXME: Is removing spaces and strip v the right thing to do?
        return remove_spaces(string).rstrip("v ").strip()

    @classmethod
    def build_value(self, string):
        """
        Return a wrapped version "value" object for a version ``string``.
        Subclasses can override. The default is a no-op and returns the string
        as-is, and is called by default at init time with the computed
        normalized_string.
        """
        return string

    def satisfies(self, constraint):
        """
        Return True is this Version satisfies the ``constraint``
        VersionConstraint. Satisfying means that this version is "within" the
        ``constraint``.
        """
        return self in constraint

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.value.__eq__(other.value)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.value.__lt__(other.value)

    def __gt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.value.__gt__(other.value)

    def __le__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.value.__le__(other.value)

    def __ge__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.value.__ge__(other.value)


@attr.s(frozen=True, order=False, hash=True)
class GenericVersion(Version):
    @classmethod
    def is_valid(cls, string):
        # generic implementation ...
        # TODO: Should use
        # https://github.com/repology/libversion/blob/master/doc/ALGORITHM.md#core-algorithm
        #  Version is split into separate all-alphabetic or all-numeric
        #  components.
        # All other characters are treated as separators. Empty components are
        # not generated.
        #   10.2alpha3..patch.4. → 10, 2, alpha, 3, patch, 4
        return super(GenericVersion, cls).is_valid(string)


@attr.s(frozen=True, order=False, eq=False, hash=True)
class PypiVersion(Version):
    """
    Python PEP 440 version as implemented in packaging with fallback to "legacy"
    """

    # TODO: ensure we deal with triple equal

    @classmethod
    def build_value(cls, string):
        """
        Return a packaging.version.LegacyVersion or packaging.version.Version
        """
        return packaging_version.Version(string)

    @classmethod
    def is_valid(cls, string):
        try:
            # Note: we consider only modern pep440 versions as valid. legacy
            # will fail validation for now.
            cls.build_value(string)
            return True
        except packaging_version.InvalidVersion:
            return False

        return False


@attr.s(frozen=True, order=False, eq=False, hash=True)
class SemverVersion(Version):
    """
    Strict semver v2.0 with 3 segments.
    """

    @classmethod
    def build_value(cls, string):
        return semantic_version.Version.coerce(string)

    @classmethod
    def is_valid(cls, string):
        try:
            cls.build_value(string)
            return True
        except ValueError:
            return False


@attr.s(frozen=True, order=False, eq=False, hash=True)
class RubygemsVersion(Version):
    """
    Rubygems encourages semver version but does not enforce it.
    Rubygems supports 4 or more segments in versions such
    as with https://rubygems.org/gems/rails/versions/5.0.0.1
    """

    @classmethod
    def build_value(cls, string):
        return gem.GemVersion(string)

    @classmethod
    def is_valid(cls, string):
        return gem.GemVersion.is_correct(string)


@attr.s(frozen=True, order=False, eq=False, hash=True)
class ArchLinuxVersion(Version):
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return arch.vercmp(self.value, other.value) == 0

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return arch.vercmp(self.value, other.value) < 0

    def __gt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return arch.vercmp(self.value, other.value) > 0

    def __le__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return arch.vercmp(self.value, other.value) <= 0

    def __ge__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return arch.vercmp(self.value, other.value) >= 0


@attr.s(frozen=True, order=False, eq=False, hash=True)
class DebianVersion(Version):
    @classmethod
    def build_value(cls, string):
        return debian.Version.from_string(string)

    @classmethod
    def is_valid(cls, string):
        return debian.Version.is_valid(string)


@attr.s(frozen=True, order=False, eq=False, hash=True)
class MavenVersion(Version):
    # See https://maven.apache.org/enforcer/enforcer-rules/versionRanges.html
    # https://github.com/apache/maven/tree/master/maven-artifact/src/main/java/org/apache/maven/artifact/versioning

    @classmethod
    def build_value(cls, string):
        return maven.Version(string)


@attr.s(frozen=True, order=False, eq=False, hash=True)
class NugetVersion(SemverVersion):
    # See https://docs.microsoft.com/en-us/nuget/concepts/package-versioning
    pass


@attr.s(frozen=True, order=False, eq=False, hash=True)
class RpmVersion(Version):
    """
    Represent an RPM version.

    For example::

    # 1:1.1.4|>=2.8.16|<=2.8.16-z
    """

    @classmethod
    def build_value(cls, string):
        return rpm.RpmVersion.from_string(string)


@total_ordering
@attr.s(frozen=True, order=False, eq=False, hash=True)
class GentooVersion(Version):
    @classmethod
    def is_valid(cls, string):
        return gentoo.is_valid(string)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return gentoo.vercmp(self.value, other.value) == 0

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return gentoo.vercmp(self.value, other.value) == -1


@attr.s(frozen=True, order=False, eq=False, hash=True)
class AlpineLinuxVersion(Version):
    @classmethod
    def is_valid(cls, string):
        return is_valid_alpine_version(string) and gentoo.is_valid(string)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return gentoo.vercmp(self.value, other.value) == 0

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return gentoo.vercmp(self.value, other.value) < 0

    def __gt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return gentoo.vercmp(self.value, other.value) > 0


@attr.s(frozen=True, order=False, eq=False, hash=True)
class LegacyOpensslVersion(Version):
    """
    Represent an Legacy Openssl Version.

    For example::
    >>> LegacyOpensslVersion("1.0.1f")
    LegacyOpensslVersion(string='1.0.1f')
    >>> LegacyOpensslVersion("1.0.2ac")
    LegacyOpensslVersion(string='1.0.2ac')
    >>> LegacyOpensslVersion("1.0.2a")
    LegacyOpensslVersion(string='1.0.2a')
    >>> LegacyOpensslVersion("3.0.2")
    Traceback (most recent call last):
        ...
    univers.versions.InvalidVersion: '3.0.2' is not a valid <class 'univers.versions.LegacyOpensslVersion'>
    """

    @classmethod
    def is_valid(cls, string):
        return bool(cls.parse(string))

    @classmethod
    def parse(cls, string):
        """
        Return a four-tuple of (major, minor, build, patch) version segments where
        major, minor, build are integers and patch is a string possibly empty.
        Return False if this is not a valid LegacyOpensslVersion.

        For example::
        >>> LegacyOpensslVersion.parse("1.0.1f")
        (1, 0, 1, 'f')
        >>> LegacyOpensslVersion.parse("1.0.2ac")
        (1, 0, 2, 'ac')
        >>> LegacyOpensslVersion.parse("2.0.2az")
        False
        """

        # All known legacy base OpenSSL versions
        all_legacy_base = (
            "0.9.1",
            "0.9.2",
            "0.9.3",
            "0.9.4",
            "0.9.5",
            "0.9.6",
            "0.9.7",
            "0.9.8",
            "1.0.0",
            "1.0.1",
            "1.0.2",
            "1.1.0",
            "1.1.1",
        )
        if not string.startswith(all_legacy_base):
            return False

        segments = string.split(".")
        if not len(segments) == 3:
            return False
        major, minor, build = segments
        major = int(major)
        minor = int(minor)
        if build.isdigit():
            build = int(build)
            patch = ""
        else:
            patch = build[1:]
            build = int(build[0])
            if patch[0].isdigit():
                return False
        return major, minor, build, patch

    @classmethod
    def build_value(cls, string):
        return cls.parse(string)

    def __str__(self):
        return f"{self.value[0]}.{self.value[1]}.{self.value[2]}{self.value[3]}"


@attr.s(frozen=True, order=False, eq=False, hash=True)
class OpensslVersion(Version):
    """
    Internally tracks two types of openssl versions
        - LegacyOpensslVersion for versions before version 3.0.0 such as 1.0.1g
        - Semver for versions from 3.0.0 and up
    For example::
    >>> old = OpensslVersion("1.1.0f")
    >>> new = OpensslVersion("3.0.1")
    >>> assert old == OpensslVersion(string="1.1.0f")
    >>> assert new == OpensslVersion(string="3.0.1")
    >>> assert old.value == LegacyOpensslVersion(string="1.1.0f")
    >>> assert new.value == SemverVersion(string="3.0.1")
    >>> OpensslVersion("1.2.4fg")
    Traceback (most recent call last):
        ...
    univers.versions.InvalidVersion: '1.2.4fg' is not a valid <class 'univers.versions.OpensslVersion'>
    """

    @classmethod
    def is_valid(cls, string):
        return cls.is_valid_new(string) or cls.is_valid_legacy(string)

    @classmethod
    def build_value(cls, string):
        """
        Return a wrapped version "value" object depending on
        whether version is legacy or semver.
        """
        if cls.is_valid_legacy(string):
            return LegacyOpensslVersion(string)
        if cls.is_valid_new(string):
            return SemverVersion(string)

    @classmethod
    def is_valid_new(cls, string):
        """
        Check the validity of new Openssl Version.

        For example::
        >>> OpensslVersion.is_valid_new("1.0.1f")
        False
        >>> OpensslVersion.is_valid_new("3.0.0")
        True
        >>> OpensslVersion.is_valid_new("3.0.2")
        True
        """
        if SemverVersion.is_valid(string):
            sem = semantic_version.Version.coerce(string)
            return sem.major >= 3

    @classmethod
    def is_valid_legacy(cls, string):
        return LegacyOpensslVersion.is_valid(string)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if not isinstance(other.value, self.value.__class__):
            return NotImplemented
        return self.value.__eq__(other.value)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if isinstance(other.value, self.value.__class__):
            return self.value.__lt__(other.value)
        # By construction legacy version is always behind Semver
        return isinstance(self.value, LegacyOpensslVersion)

    def __gt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if isinstance(other.value, self.value.__class__):
            return self.value.__gt__(other.value)
        # By construction semver version is always ahead of legacy
        return isinstance(self.value, SemverVersion)

    def __le__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if isinstance(other.value, self.value.__class__):
            return self.value.__le__(other.value)
        # version value are of diff type, then legacy one is always behind semver
        return isinstance(self.value, LegacyOpensslVersion)

    def __ge__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if isinstance(other.value, self.value.__class__):
            return self.value.__ge__(other.value)
        # version value are of diff type, then semver one is always ahead of legacy
        return isinstance(self.value, SemverVersion)



#
# Copyright (c) 2006-2019, pkgcore contributors
# SPDX-License-Identifier: BSD-3-Clause
# Version comparison utility extracted from pkgcore and further stripped down.
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

from random import shuffle

from univers.gentoo import parse_version_and_revision
from univers.gentoo import vercmp


def generate_misc_sufs():
    simple_good_sufs = ["_alpha", "_beta", "_pre", "_p"]
    suf_nums = list(range(100))
    shuffle(suf_nums)

    good_sufs = simple_good_sufs + [f"{x}{suf_nums.pop()}" for x in simple_good_sufs]

    l = len(good_sufs)
    good_sufs = good_sufs + [good_sufs[x] + good_sufs[l - x - 1] for x in range(l)]

    bad_sufs = ["_a", "_9", "_"] + [x + " " for x in simple_good_sufs]
    return good_sufs, bad_sufs


good_vers = ("1", "2.3.4", "2.3.4a", "02.3", "2.03", "3d", "3D")

good_sufs, bad_sufs = generate_misc_sufs()
good_revs = ("-r1", "-r300", "-r0", "", "-r1000000000000000000")


def older(v1, v2):
    return vercmp(v1, v2) < 0


def newer(v1, v2):
    return vercmp(v1, v2) > 0


def equal(v1, v2):
    return vercmp(v1, v2) == 0


class TestCPV:
    def test_parse_version_and_revision_rev(self):

        for x in (10, 18, 19, 36, 100):
            _ver, rev = parse_version_and_revision(f"1-r0{'0' * x}")
            assert rev == 0

            _ver, rev = parse_version_and_revision(f"1-r1{'0' * x}1")
            assert int(rev) == int(f"1{'0' * x}1")

    def test_parse_version_and_revisions(self):
        for ver in good_vers:
            v_r = parse_version_and_revision(ver)
            assert v_r == (ver, 0)
            for rev in good_revs:
                version = f"{ver}{rev}"
                try:
                    v_r = parse_version_and_revision(version)
                    expected_rev = rev.lstrip("-r")
                    expected_rev = expected_rev and int(expected_rev) or 0
                    assert v_r == (ver, expected_rev), f"Failed to parse: {version}"
                except Exception as e:
                    raise Exception(f"Failed to parse: {version!r}") from e

                for suf in good_sufs:
                    version = f"{ver}{suf}{rev}"
                    v_r = parse_version_and_revision(version)
                    assert v_r == (f"{ver}{suf}", expected_rev)

    def test_cmp(self):
        assert older("0.1", "0.2")

    def test_cmp1(self):
        base = "0.7.1"
        base = equal(base, base)
        for rev in ("", "-r1"):
            last = None
            for suf in ["_alpha", "_beta", "_pre", "", "_p"]:
                if suf == "":
                    sufs = [suf]
                else:
                    sufs = [suf, f"{suf}4"]
                for x in sufs:
                    cur = f"{base}{x}{rev}"
                    assert equal(cur, f"{base}{x}{rev}")
                    if last is not None:
                        assert newer(cur, last)

    def test_cmp2(self):
        assert newer("6a", "6")
        assert newer("6a-r1", "6a")
        assert newer("6.0", "6")
        assert newer("6.0.0", "6.0b")
        assert newer("6.02", "6.0.0")

    def test_cmp_float(self):
        # float comparison rules.
        assert newer("6.2", "6.054")
        assert equal("6", "6")
        assert newer("6.0_alpha0_p1", "6.0_alpha")
        assert equal("6.0_alpha", "6.0_alpha0")
        assert newer("6.1", "6.09")
        assert newer("6.0.1", "6.0")
        assert newer("12.2.5", "12.2b")

    def test_cmp_287848(self):
        # test for gentoo bug 287848
        assert newer("12.2.5", "12.2b")
        assert newer("12.2.5-r1", "12.2b")

    def test_cmp_equiv(self):
        # equivalent versions
        assert equal("6.01.0", "6.010.0")
        assert equal("6.0.1", "6.000.1")

    def test_cmp_equiv_rev(self):
        # equivalent revisions
        assert equal("6.01.0", "6.01.0-r0")
        assert equal("6.01.0-r0", "6.01.0-r00")
        assert equal("6.01.0-r1", "6.01.0-r001")

        for v1, v2 in (
            ("1.001000000000000000001", "1.001000000000000000002"),
            ("1.00100000000", "1.0010000000000000001"),
            ("1.01", "1.1"),
        ):
            assert newer(f"{v2}", f"{v1}")

        for x in (18, 36, 100):
            s = "0" * x
            assert newer(f"10{s}1", f"1{s}1")

        for x in (18, 36, 100):
            s = "0" * x
            assert newer(f"1-r10{s}1", f"1-r1{s}1")

        assert newer("1.60_p2010081516093", "1.60_p2009072801401")

        assert newer("1.60_p20100815160931", "1.60_p20090728014017")

        assert newer("1.60_p20100815160931", "1.60_p20090728014017-r1")

        # Regression test: python does comparison slightly differently
        # if the classes do not match exactly (it prefers rich
        # comparison over __cmp__).

        assert not equal("6.0_alpha0_p1", "6.0_alpha")
        assert equal("6.0_alpha0", "6.0_alpha")

        assert not equal("6.0", "")
        assert equal("6.0", "6.0-r0")

    def test_r0_revisions_with_single_0(self):
        ver, rev = parse_version_and_revision("1.0-r0")
        assert ver == "1.0"
        assert rev == 0

    def test_r0_revisions_with_multiple_0(self):
        ver, rev = parse_version_and_revision("1.0-r000")
        assert ver == "1.0"
        assert rev == 0

    def test_r0_revisions_with_single_0_prefix(self):
        ver, rev = parse_version_and_revision("1.0-r01")
        assert ver == "1.0"
        assert rev == 1

    def test_r0_revisions_with_multiple_0_prefix(self):
        ver, rev = parse_version_and_revision("11.0-r0001")
        assert ver == "11.0"
        assert rev == 1



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import operator
from typing import NamedTuple

import pytest
from univers.versions import AlpineLinuxVersion
from univers.versions import InvalidVersion


class AlpineVersionComparison(NamedTuple):
    """
    We have this class to drive the tests, where we compare
    `version1` and `version2` on the basis of the `comparator`
    """

    version1: str
    comparator: str
    version2: str

    @classmethod
    def from_string(cls, s: str):
        segments = s.split()

        assert len(segments) == 3
        return cls(*segments)

    def compare(self):
        """
        Compare version1 and version2 with comparator
        and return True if comparison is valid, False otherwise
        """
        v1 = AlpineLinuxVersion(self.version1)
        v2 = AlpineLinuxVersion(self.version2)
        COMPARATORS = {
            "<": operator.lt,
            ">": operator.gt,
            "=": operator.eq,
        }
        comparator = COMPARATORS[self.comparator]
        return comparator(v1, v2)


@pytest.mark.parametrize("test_case", open("./tests/data/alpine_test.txt").read().splitlines(False))
def test_alpine_vers_cmp(test_case):
    avc = AlpineVersionComparison.from_string(test_case)
    assert avc.compare()


@pytest.mark.parametrize(
    "test_case",
    [
        # these are the tests are not supported yet
        # when we start supporting these version,
        # they will be moved back to main test suite
        "2.10.1 > 02.08.01b",
        "02.08.01b < 4.77",
        "23_foo > 4_beta",
        "1.06-r6 < 006",
        "006 > 1.0.0",
        "2.10.1 > 02.08.01b",
        "02.08.01b < 4.77",
        "2.2.3-r2 < 013",
        "013 < 014-r1",
        "014-r1 > 1.3.1-r1",
        "3.0.0-r2 < 021109-r3",
        "021109-r3 < 20060512",
        "0.9.28.1 < 087-r1",
        "087-r1 < 103",
        # invalid. do string sort
        "1.0 < 1.0bc",
    ],
)
def test_invalid_alpine_vers_cmp(test_case):
    avc = AlpineVersionComparison.from_string(test_case)
    with pytest.raises(InvalidVersion):
        AlpineLinuxVersion(avc.version1)
        AlpineLinuxVersion(avc.version2)



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import pytest

from univers import versions
from univers.version_constraint import VersionConstraint


@pytest.mark.parametrize(
    "version, spec, expected",
    [
        ("2.7", "<=3.4", True),
        ("2.7.1", "<=3.4", True),
        ("2.7.1rc1", "<=3.4", True),
        ("2.7.15", "<=3.4", True),
        ("2.7.15rc1", "<=3.4", True),
        ("2.7", ">=3.4", False),
        ("2.7.1", ">=3.4", False),
        ("2.7.1rc1", ">=3.4", False),
        ("2.7.15", ">=3.4", False),
        ("2.7.15rc1", ">=3.4", False),
        ("0.0.0", ">=1.0.0", False),
        ("1.2.3", ">=1.0.0", True),
        ("1.2.3b1", ">=1.0.0", True),
        ("1.0.1b1", ">=1.0.0", True),
        ("1.0.0b1", ">=1.0.0", False),
        ("1.0.0b1", ">=1.0.0b1", True),
    ],
)
def test_pypi_comparison(version, spec, expected):
    version = versions.PypiVersion(version)
    constraint = VersionConstraint.from_string(
        string=spec,
        version_class=versions.PypiVersion,
    )
    assert (version in constraint) is expected


@pytest.mark.parametrize(
    "version, spec, expected",
    [
        ("2.7.1", "<=3.4.3", True),
        ("1.1.0", ">1.0.0", True),
        ("2.0.0", "<=2.0.0", True),
        ("1.9999.9999", "<=2.0.0", True),
        ("0.2.9", "<=2.0.0", True),
        ("1.9999.9999", "<2.0.0", True),
        ("0.1.1-alpha", ">=0.1.1-beta", False),
        ("1.0.0+20130313144700", "=1.0.0+9999999999", False),
    ],
)
def test_semver_comparison(version, spec, expected):
    version = versions.SemverVersion(version)
    constraint = VersionConstraint.from_string(
        string=spec,
        version_class=versions.SemverVersion,
    )
    assert (version in constraint) is expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) nexB Inc. and others. All rights reserved.
# ScanCode is a trademark of nexB Inc.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/skeleton for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import itertools
import os
import sys

import click

import utils_thirdparty
import utils_requirements

TRACE = True


@click.command()
@click.option(
    "-r",
    "--requirements",
    "requirements_files",
    type=click.Path(exists=True, readable=True, path_type=str, dir_okay=False),
    metavar="REQUIREMENT-FILE",
    multiple=True,
    required=False,
    help="Path to pip requirements file(s) listing thirdparty packages.",
)
@click.option(
    "--spec",
    "--specifier",
    "specifiers",
    type=str,
    metavar="SPECIFIER",
    multiple=True,
    required=False,
    help="Thirdparty package name==version specification(s) as in django==1.2.3. "
    "With --latest-version a plain package name is also acceptable.",
)
@click.option(
    "-l",
    "--latest-version",
    is_flag=True,
    help="Get the latest version of all packages, ignoring any specified versions.",
)
@click.option(
    "-d",
    "--dest",
    "dest_dir",
    type=click.Path(exists=True, readable=True, path_type=str, file_okay=False),
    metavar="DIR",
    default=utils_thirdparty.THIRDPARTY_DIR,
    show_default=True,
    help="Path to the detsination directory where to save downloaded wheels, "
    "sources, ABOUT and LICENSE files..",
)
@click.option(
    "-w",
    "--wheels",
    is_flag=True,
    help="Download wheels.",
)
@click.option(
    "-s",
    "--sdists",
    is_flag=True,
    help="Download source sdists tarballs.",
)
@click.option(
    "-p",
    "--python-version",
    "python_versions",
    type=click.Choice(utils_thirdparty.PYTHON_VERSIONS),
    metavar="PYVER",
    default=utils_thirdparty.PYTHON_VERSIONS,
    show_default=True,
    multiple=True,
    help="Python version(s) to use for wheels.",
)
@click.option(
    "-o",
    "--operating-system",
    "operating_systems",
    type=click.Choice(utils_thirdparty.PLATFORMS_BY_OS),
    metavar="OS",
    default=tuple(utils_thirdparty.PLATFORMS_BY_OS),
    multiple=True,
    show_default=True,
    help="OS(ses) to use for wheels: one of linux, mac or windows.",
)
@click.option(
    "--index-url",
    "index_urls",
    type=str,
    metavar="INDEX",
    default=utils_thirdparty.PYPI_INDEXES,
    show_default=True,
    multiple=True,
    help="PyPI index URL(s) to use for wheels and sources, in order of preferences.",
)
@click.help_option("-h", "--help")
def fetch_thirdparty(
    requirements_files,
    specifiers,
    latest_version,
    dest_dir,
    python_versions,
    operating_systems,
    wheels,
    sdists,
    index_urls,
):
    """
    Download to --dest-dir THIRDPARTY_DIR the PyPI wheels, source distributions,
    and their ABOUT metadata, license and notices files.

    Download the PyPI packages listed in the combination of:
    - the pip requirements --requirements REQUIREMENT-FILE(s),
    - the pip name==version --specifier SPECIFIER(s)
    - any pre-existing wheels or sdsists found in --dest-dir THIRDPARTY_DIR.

    Download wheels with the --wheels option for the ``--python-version`` PYVER(s)
    and ``--operating_system`` OS(s) combinations defaulting to all supported combinations.

    Download sdists tarballs with the --sdists option.

    Generate or Download .ABOUT, .LICENSE and .NOTICE files for all the wheels and sources fetched.

    Download wheels and sdists the provided PyPI simple --index-url INDEX(s) URLs.
    """
    print(f"COLLECTING REQUIRED NAMES & VERSIONS FROM {dest_dir}")
    existing_packages_by_nv = {
        (package.name, package.version): package
        for package in utils_thirdparty.get_local_packages(directory=dest_dir)
    }

    required_name_versions = set(existing_packages_by_nv.keys())

    for req_file in requirements_files:
        nvs = utils_requirements.load_requirements(
            requirements_file=req_file,
            with_unpinned=latest_version,
        )
        required_name_versions.update(nvs)

    for specifier in specifiers:
        nv = utils_requirements.get_name_version(
            requirement=specifier,
            with_unpinned=latest_version,
        )
        required_name_versions.add(nv)

    if not required_name_versions:
        print("Error: no requirements requested.")
        sys.exit(1)

    if not os.listdir(dest_dir) and not (wheels or sdists):
        print("Error: one or both of --wheels  and --sdists is required.")
        sys.exit(1)

    if latest_version:
        latest_name_versions = set()
        names = set(name for name, _version in sorted(required_name_versions))
        for name in sorted(names):
            latests = utils_thirdparty.PypiPackage.sorted(
                utils_thirdparty.get_package_versions(
                    name=name, version=None, index_urls=index_urls
                )
            )
            if not latests:
                print(f"No distribution found for: {name}")
            continue
            latest = latests[-1]
            latest_name_versions.add((latest.name, latest.version))
        required_name_versions = latest_name_versions

    if TRACE:
        print("required_name_versions:", required_name_versions)

    if wheels:
        # create the environments matrix we need for wheels
        evts = itertools.product(python_versions, operating_systems)
        environments = [utils_thirdparty.Environment.from_pyver_and_os(pyv, os) for pyv, os in evts]

    wheels_not_found = {}
    sdists_not_found = {}
    # iterate over requirements, one at a time
    for name, version in sorted(required_name_versions):
        nv = name, version
        existing_package = existing_packages_by_nv.get(nv)
        if wheels:
            for environment in environments:
                if existing_package:
                    existing_wheels = list(
                        existing_package.get_supported_wheels(environment=environment)
                    )
                else:
                    existing_wheels = None

                if existing_wheels:
                    if TRACE:
                        print(
                            f"====> Wheels already available: {name}=={version} on: {environment}: {existing_package.wheels!r}"
                        )
                    if all(w.is_pure() for w in existing_wheels):
                        break
                    else:
                        continue

                if TRACE:
                    print(f"Fetching wheel for: {name}=={version} on: {environment}")

                try:
                    (
                        fetched_wheel_filenames,
                        existing_wheel_filenames,
                    ) = utils_thirdparty.download_wheel(
                        name=name,
                        version=version,
                        environment=environment,
                        dest_dir=dest_dir,
                        index_urls=index_urls,
                    )
                    if TRACE:
                        if existing_wheel_filenames:
                            print(
                                f"    ====> Wheels already available: {name}=={version} on: {environment}"
                            )
                            for whl in existing_wheel_filenames:
                                print(f"        {whl}")
                        if fetched_wheel_filenames:
                            print(f"    ====> Wheels fetched: {name}=={version} on: {environment}")
                            for whl in fetched_wheel_filenames:
                                print(f"        {whl}")

                    fwfns = fetched_wheel_filenames + existing_wheel_filenames

                    if all(utils_thirdparty.Wheel.from_filename(f).is_pure() for f in fwfns):
                        break

                except utils_thirdparty.DistributionNotFound as e:
                    wheels_not_found[f"{name}=={version}"] = str(e)

        if sdists:
            if existing_package and existing_package.sdist:
                if TRACE:
                    print(
                        f"  ====> Sdist already available: {name}=={version}: {existing_package.sdist!r}"
                    )
                continue

            if TRACE:
                print(f"  Fetching sdist for: {name}=={version}")

            try:
                fetched = utils_thirdparty.download_sdist(
                    name=name,
                    version=version,
                    dest_dir=dest_dir,
                    index_urls=index_urls,
                )

                if TRACE:
                    if not fetched:
                        print(
                            f"    ====> Sdist already available: {name}=={version} on: {environment}"
                        )
                    else:
                        print(
                            f"    ====> Sdist fetched: {fetched} for {name}=={version} on: {environment}"
                        )

            except utils_thirdparty.DistributionNotFound as e:
                sdists_not_found[f"{name}=={version}"] = str(e)

    if wheels and wheels_not_found:
        print(f"==> MISSING WHEELS")
        for wh in wheels_not_found:
            print(f"  {wh}")

    if sdists and sdists_not_found:
        print(f"==> MISSING SDISTS")
        for sd in sdists_not_found:
            print(f"  {sd}")

    print(f"==> FETCHING OR CREATING ABOUT AND LICENSE FILES")
    utils_thirdparty.fetch_abouts_and_licenses(dest_dir=dest_dir)
    utils_thirdparty.clean_about_files(dest_dir=dest_dir)

    # check for problems
    print(f"==> CHECK FOR PROBLEMS")
    utils_thirdparty.find_problems(
        dest_dir=dest_dir,
        report_missing_sources=sdists,
        report_missing_wheels=wheels,
    )


if __name__ == "__main__":
    fetch_thirdparty()



# Copyright (c) 2009-2020 by Pacman Development Team
# Copyright (c) 2008 by Dan McGee <dan@archlinux.org>
# SPDX-License-Identifier: Apache-2.0
# this has been significantly modified from the original
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

from univers.versions import ArchLinuxVersion


def test_same_length():
    assert ArchLinuxVersion("1.5.0") == ArchLinuxVersion("1.5.0")
    assert ArchLinuxVersion("1.5.1") > ArchLinuxVersion("1.5.0")


def test_mixed_length():
    assert ArchLinuxVersion("1.5.1") > ArchLinuxVersion("1.5")


def test_with_pkgrel_same_length():
    assert ArchLinuxVersion("1.5.0-1") == ArchLinuxVersion("1.5.0-1")
    assert ArchLinuxVersion("1.5.0-1") < ArchLinuxVersion("1.5.0-2")
    assert ArchLinuxVersion("1.5.0-1") < ArchLinuxVersion("1.5.1-1")
    assert ArchLinuxVersion("1.5.0-2") < ArchLinuxVersion("1.5.1-1")


def test_alpha_dotted_versions():
    assert ArchLinuxVersion("1.5.a") > ArchLinuxVersion("1.5")
    assert ArchLinuxVersion("1.5.b") > ArchLinuxVersion("1.5.a")
    assert ArchLinuxVersion("1.5.1") > ArchLinuxVersion("1.5.b")


def test_with_epoch():
    assert ArchLinuxVersion("0:1.0") == ArchLinuxVersion("0:1.0")
    assert ArchLinuxVersion("0:1.0") < ArchLinuxVersion("0:1.1")
    assert ArchLinuxVersion("1:1.0") > ArchLinuxVersion("0:1.0")
    assert ArchLinuxVersion("1:1.0") > ArchLinuxVersion("0:1.1")
    assert ArchLinuxVersion("1:1.0") < ArchLinuxVersion("2:1.1")


def test_with_epoch_mixed_pkgrel():
    assert ArchLinuxVersion("1:1.0") > ArchLinuxVersion("0:1.0-1")
    assert ArchLinuxVersion("1:1.0-1") > ArchLinuxVersion("0:1.1-1")


def test_with_only_one_version_with_epoch():
    assert ArchLinuxVersion("0:1.0") == ArchLinuxVersion("1.0")
    assert ArchLinuxVersion("0:1.0") < ArchLinuxVersion("1.1")
    assert ArchLinuxVersion("0:1.1") > ArchLinuxVersion("1.0")
    assert ArchLinuxVersion("1:1.0") > ArchLinuxVersion("1.0")
    assert ArchLinuxVersion("1:1.0") > ArchLinuxVersion("1.1")
    assert ArchLinuxVersion("1:1.1") > ArchLinuxVersion("1.1")


def test_alpha_dot_and_dashes():
    assert ArchLinuxVersion("1.5.b-1") == ArchLinuxVersion("1.5.b")
    assert ArchLinuxVersion("1.5-1") < ArchLinuxVersion("1.5.b")


# def test_same_content_different_separators():
#     assert ArchLinuxVersion("2.0") == ArchLinuxVersion("2_0")
#     assert ArchLinuxVersion("2.0_a") == ArchLinuxVersion("2_0.a")
#     assert ArchLinuxVersion("2.0a ") < ArchLinuxVersion("2.0.a")
#     assert ArchLinuxVersion("2___a") == ArchLinuxVersion("2_a")


def test_with_pkgrel_mixed_length():
    assert ArchLinuxVersion("1.5-1") < ArchLinuxVersion("1.5.1-1")
    assert ArchLinuxVersion("1.5-2") < ArchLinuxVersion("1.5.1-1")
    assert ArchLinuxVersion("1.5-2") < ArchLinuxVersion("1.5.1-2")


def test_with_mixed_pkgrel_inclusion():
    assert ArchLinuxVersion("1.5") == ArchLinuxVersion("1.5-1")
    assert ArchLinuxVersion("1.5-1") == ArchLinuxVersion("1.5")
    assert ArchLinuxVersion("1.1-1") == ArchLinuxVersion("1.1")
    assert ArchLinuxVersion("1.0-1") < ArchLinuxVersion("1.1")
    assert ArchLinuxVersion("1.1-1") > ArchLinuxVersion("1.0")


def test_alphanumeric_versions():
    assert ArchLinuxVersion("1.5b-1") < ArchLinuxVersion("1.5-1")
    assert ArchLinuxVersion("1.5b  ") < ArchLinuxVersion("1.5  ")
    assert ArchLinuxVersion("1.5b-1") < ArchLinuxVersion("1.5  ")
    assert ArchLinuxVersion("1.5b  ") < ArchLinuxVersion("1.5.1")


def test_manpage_cases():
    assert ArchLinuxVersion("1.0a") < ArchLinuxVersion("1.0alpha")
    assert ArchLinuxVersion("1.0alpha") < ArchLinuxVersion("1.0b")
    assert ArchLinuxVersion("1.0b") < ArchLinuxVersion("1.0beta")
    assert ArchLinuxVersion("1.0beta") < ArchLinuxVersion("1.0rc")
    assert ArchLinuxVersion("1.0rc") < ArchLinuxVersion("1.0")
